# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: exchange.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0e\x65xchange.proto\"G\n\x0f\x45xchangeAddress\x12\x11\n\tcoin_type\x18\x01 \x01(\t\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\x12\x10\n\x08\x64\x65st_tag\x18\x03 \x01(\t\"\xcc\x02\n\x12\x45xchangeResponseV2\x12)\n\x0f\x64\x65posit_address\x18\x01 \x01(\x0b\x32\x10.ExchangeAddress\x12\x16\n\x0e\x64\x65posit_amount\x18\x02 \x01(\x0c\x12\x12\n\nexpiration\x18\x03 \x01(\x03\x12\x13\n\x0bquoted_rate\x18\x04 \x01(\x0c\x12,\n\x12withdrawal_address\x18\x05 \x01(\x0b\x32\x10.ExchangeAddress\x12\x19\n\x11withdrawal_amount\x18\x06 \x01(\x0c\x12(\n\x0ereturn_address\x18\x07 \x01(\x0b\x32\x10.ExchangeAddress\x12\x0f\n\x07\x61pi_key\x18\x08 \x01(\x0c\x12\x11\n\tminer_fee\x18\t \x01(\x0c\x12\x10\n\x08order_id\x18\n \x01(\x0c\x12!\n\x04type\x18\x0b \x01(\x0e\x32\n.OrderType:\x07Precise\"T\n\x16SignedExchangeResponse\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\'\n\nresponseV2\x18\x03 \x01(\x0b\x32\x13.ExchangeResponseV2*#\n\tOrderType\x12\x0b\n\x07Precise\x10\x00\x12\t\n\x05Quick\x10\x01\x42.\n\x1b\x63om.keepkey.device-protocolB\x0fKeepKeyExchange')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'exchange_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\033com.keepkey.device-protocolB\017KeepKeyExchange'
  _ORDERTYPE._serialized_start=512
  _ORDERTYPE._serialized_end=547
  _EXCHANGEADDRESS._serialized_start=18
  _EXCHANGEADDRESS._serialized_end=89
  _EXCHANGERESPONSEV2._serialized_start=92
  _EXCHANGERESPONSEV2._serialized_end=424
  _SIGNEDEXCHANGERESPONSE._serialized_start=426
  _SIGNEDEXCHANGERESPONSE._serialized_end=510
# @@protoc_insertion_point(module_scope)
