# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: messages.proto
"""Generated protocol buffer code."""
from google.protobuf.internal import builder as _builder
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import types_pb2 as types__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x0emessages.proto\x1a\x0btypes.proto\"\x0c\n\nInitialize\"\r\n\x0bGetFeatures\"\xf0\x03\n\x08\x46\x65\x61tures\x12\x0e\n\x06vendor\x18\x01 \x01(\t\x12\x15\n\rmajor_version\x18\x02 \x01(\r\x12\x15\n\rminor_version\x18\x03 \x01(\r\x12\x15\n\rpatch_version\x18\x04 \x01(\r\x12\x17\n\x0f\x62ootloader_mode\x18\x05 \x01(\x08\x12\x11\n\tdevice_id\x18\x06 \x01(\t\x12\x16\n\x0epin_protection\x18\x07 \x01(\x08\x12\x1d\n\x15passphrase_protection\x18\x08 \x01(\x08\x12\x10\n\x08language\x18\t \x01(\t\x12\r\n\x05label\x18\n \x01(\t\x12\x18\n\x05\x63oins\x18\x0b \x03(\x0b\x32\t.CoinType\x12\x13\n\x0binitialized\x18\x0c \x01(\x08\x12\x10\n\x08revision\x18\r \x01(\x0c\x12\x17\n\x0f\x62ootloader_hash\x18\x0e \x01(\x0c\x12\x10\n\x08imported\x18\x0f \x01(\x08\x12\x12\n\npin_cached\x18\x10 \x01(\x08\x12\x19\n\x11passphrase_cached\x18\x11 \x01(\x08\x12\x1d\n\x08policies\x18\x12 \x03(\x0b\x32\x0b.PolicyType\x12\r\n\x05model\x18\x15 \x01(\t\x12\x18\n\x10\x66irmware_variant\x18\x16 \x01(\t\x12\x15\n\rfirmware_hash\x18\x17 \x01(\x0c\x12\x11\n\tno_backup\x18\x18 \x01(\x08\"*\n\x0cGetCoinTable\x12\r\n\x05start\x18\x01 \x01(\r\x12\x0b\n\x03\x65nd\x18\x02 \x01(\r\"L\n\tCoinTable\x12\x18\n\x05table\x18\x01 \x03(\x0b\x32\t.CoinType\x12\x11\n\tnum_coins\x18\x02 \x01(\r\x12\x12\n\nchunk_size\x18\x03 \x01(\r\"\x0e\n\x0c\x43learSession\"y\n\rApplySettings\x12\x10\n\x08language\x18\x01 \x01(\t\x12\r\n\x05label\x18\x02 \x01(\t\x12\x16\n\x0euse_passphrase\x18\x03 \x01(\x08\x12\x1a\n\x12\x61uto_lock_delay_ms\x18\x04 \x01(\r\x12\x13\n\x0bu2f_counter\x18\x05 \x01(\r\"\x1b\n\tChangePin\x12\x0e\n\x06remove\x18\x01 \x01(\x08\"i\n\x04Ping\x12\x0f\n\x07message\x18\x01 \x01(\t\x12\x19\n\x11\x62utton_protection\x18\x02 \x01(\x08\x12\x16\n\x0epin_protection\x18\x03 \x01(\x08\x12\x1d\n\x15passphrase_protection\x18\x04 \x01(\x08\"\x1a\n\x07Success\x12\x0f\n\x07message\x18\x01 \x01(\t\"6\n\x07\x46\x61ilure\x12\x1a\n\x04\x63ode\x18\x01 \x01(\x0e\x32\x0c.FailureType\x12\x0f\n\x07message\x18\x02 \x01(\t\"?\n\rButtonRequest\x12 \n\x04\x63ode\x18\x01 \x01(\x0e\x32\x12.ButtonRequestType\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\t\"\x0b\n\tButtonAck\"7\n\x10PinMatrixRequest\x12#\n\x04type\x18\x01 \x01(\x0e\x32\x15.PinMatrixRequestType\"\x1b\n\x0cPinMatrixAck\x12\x0b\n\x03pin\x18\x01 \x02(\t\"\x08\n\x06\x43\x61ncel\"\x13\n\x11PassphraseRequest\"#\n\rPassphraseAck\x12\x12\n\npassphrase\x18\x01 \x02(\t\"\x1a\n\nGetEntropy\x12\x0c\n\x04size\x18\x01 \x02(\r\"\x1a\n\x07\x45ntropy\x12\x0f\n\x07\x65ntropy\x18\x01 \x02(\x0c\"\xa2\x01\n\x0cGetPublicKey\x12\x11\n\taddress_n\x18\x01 \x03(\r\x12\x18\n\x10\x65\x63\x64sa_curve_name\x18\x02 \x01(\t\x12\x14\n\x0cshow_display\x18\x03 \x01(\x08\x12\x1a\n\tcoin_name\x18\x04 \x01(\t:\x07\x42itcoin\x12\x33\n\x0bscript_type\x18\x05 \x01(\x0e\x32\x10.InputScriptType:\x0cSPENDADDRESS\"4\n\tPublicKey\x12\x19\n\x04node\x18\x01 \x02(\x0b\x32\x0b.HDNodeType\x12\x0c\n\x04xpub\x18\x02 \x01(\t\"\xb3\x01\n\nGetAddress\x12\x11\n\taddress_n\x18\x01 \x03(\r\x12\x1a\n\tcoin_name\x18\x02 \x01(\t:\x07\x42itcoin\x12\x14\n\x0cshow_display\x18\x03 \x01(\x08\x12+\n\x08multisig\x18\x04 \x01(\x0b\x32\x19.MultisigRedeemScriptType\x12\x33\n\x0bscript_type\x18\x05 \x01(\x0e\x32\x10.InputScriptType:\x0cSPENDADDRESS\"\x1a\n\x07\x41\x64\x64ress\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x02(\t\"\x0c\n\nWipeDevice\"\xbb\x01\n\nLoadDevice\x12\x10\n\x08mnemonic\x18\x01 \x01(\t\x12\x19\n\x04node\x18\x02 \x01(\x0b\x32\x0b.HDNodeType\x12\x0b\n\x03pin\x18\x03 \x01(\t\x12\x1d\n\x15passphrase_protection\x18\x04 \x01(\x08\x12\x19\n\x08language\x18\x05 \x01(\t:\x07\x65nglish\x12\r\n\x05label\x18\x06 \x01(\t\x12\x15\n\rskip_checksum\x18\x07 \x01(\x08\x12\x13\n\x0bu2f_counter\x18\x08 \x01(\r\"\xe1\x01\n\x0bResetDevice\x12\x16\n\x0e\x64isplay_random\x18\x01 \x01(\x08\x12\x15\n\x08strength\x18\x02 \x01(\r:\x03\x32\x35\x36\x12\x1d\n\x15passphrase_protection\x18\x03 \x01(\x08\x12\x16\n\x0epin_protection\x18\x04 \x01(\x08\x12\x19\n\x08language\x18\x05 \x01(\t:\x07\x65nglish\x12\r\n\x05label\x18\x06 \x01(\t\x12\x11\n\tno_backup\x18\x07 \x01(\x08\x12\x1a\n\x12\x61uto_lock_delay_ms\x18\x08 \x01(\r\x12\x13\n\x0bu2f_counter\x18\t \x01(\r\"\x10\n\x0e\x45ntropyRequest\"\x1d\n\nEntropyAck\x12\x0f\n\x07\x65ntropy\x18\x01 \x01(\x0c\"\xff\x01\n\x0eRecoveryDevice\x12\x12\n\nword_count\x18\x01 \x01(\r\x12\x1d\n\x15passphrase_protection\x18\x02 \x01(\x08\x12\x16\n\x0epin_protection\x18\x03 \x01(\x08\x12\x19\n\x08language\x18\x04 \x01(\t:\x07\x65nglish\x12\r\n\x05label\x18\x05 \x01(\t\x12\x18\n\x10\x65nforce_wordlist\x18\x06 \x01(\x08\x12\x1c\n\x14use_character_cipher\x18\x07 \x01(\x08\x12\x1a\n\x12\x61uto_lock_delay_ms\x18\x08 \x01(\r\x12\x13\n\x0bu2f_counter\x18\t \x01(\r\x12\x0f\n\x07\x64ry_run\x18\n \x01(\x08\"\r\n\x0bWordRequest\"\x17\n\x07WordAck\x12\x0c\n\x04word\x18\x01 \x02(\t\";\n\x10\x43haracterRequest\x12\x10\n\x08word_pos\x18\x01 \x02(\r\x12\x15\n\rcharacter_pos\x18\x02 \x02(\r\"?\n\x0c\x43haracterAck\x12\x11\n\tcharacter\x18\x01 \x01(\t\x12\x0e\n\x06\x64\x65lete\x18\x02 \x01(\x08\x12\x0c\n\x04\x64one\x18\x03 \x01(\x08\"\x82\x01\n\x0bSignMessage\x12\x11\n\taddress_n\x18\x01 \x03(\r\x12\x0f\n\x07message\x18\x02 \x02(\x0c\x12\x1a\n\tcoin_name\x18\x03 \x01(\t:\x07\x42itcoin\x12\x33\n\x0bscript_type\x18\x04 \x01(\x0e\x32\x10.InputScriptType:\x0cSPENDADDRESS\"`\n\rVerifyMessage\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\x0c\x12\x0f\n\x07message\x18\x03 \x01(\x0c\x12\x1a\n\tcoin_name\x18\x04 \x01(\t:\x07\x42itcoin\"6\n\x10MessageSignature\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x11\n\tsignature\x18\x02 \x01(\x0c\"v\n\x0e\x45ncryptMessage\x12\x0e\n\x06pubkey\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\x0c\x12\x14\n\x0c\x64isplay_only\x18\x03 \x01(\x08\x12\x11\n\taddress_n\x18\x04 \x03(\r\x12\x1a\n\tcoin_name\x18\x05 \x01(\t:\x07\x42itcoin\"@\n\x10\x45ncryptedMessage\x12\r\n\x05nonce\x18\x01 \x01(\x0c\x12\x0f\n\x07message\x18\x02 \x01(\x0c\x12\x0c\n\x04hmac\x18\x03 \x01(\x0c\"Q\n\x0e\x44\x65\x63ryptMessage\x12\x11\n\taddress_n\x18\x01 \x03(\r\x12\r\n\x05nonce\x18\x02 \x01(\x0c\x12\x0f\n\x07message\x18\x03 \x01(\x0c\x12\x0c\n\x04hmac\x18\x04 \x01(\x0c\"4\n\x10\x44\x65\x63ryptedMessage\x12\x0f\n\x07message\x18\x01 \x01(\x0c\x12\x0f\n\x07\x61\x64\x64ress\x18\x02 \x01(\t\"\x8c\x01\n\x0e\x43ipherKeyValue\x12\x11\n\taddress_n\x18\x01 \x03(\r\x12\x0b\n\x03key\x18\x02 \x01(\t\x12\r\n\x05value\x18\x03 \x01(\x0c\x12\x0f\n\x07\x65ncrypt\x18\x04 \x01(\x08\x12\x16\n\x0e\x61sk_on_encrypt\x18\x05 \x01(\x08\x12\x16\n\x0e\x61sk_on_decrypt\x18\x06 \x01(\x08\x12\n\n\x02iv\x18\x07 \x01(\x0c\"!\n\x10\x43ipheredKeyValue\x12\r\n\x05value\x18\x01 \x01(\x0c\"\xce\x01\n\x06SignTx\x12\x15\n\routputs_count\x18\x01 \x02(\r\x12\x14\n\x0cinputs_count\x18\x02 \x02(\r\x12\x1a\n\tcoin_name\x18\x03 \x01(\t:\x07\x42itcoin\x12\x12\n\x07version\x18\x04 \x01(\r:\x01\x31\x12\x14\n\tlock_time\x18\x05 \x01(\r:\x01\x30\x12\x0e\n\x06\x65xpiry\x18\x06 \x01(\r\x12\x14\n\x0coverwintered\x18\x07 \x01(\x08\x12\x18\n\x10version_group_id\x18\x08 \x01(\r\x12\x11\n\tbranch_id\x18\n \x01(\r\"\x85\x01\n\tTxRequest\x12\"\n\x0crequest_type\x18\x01 \x01(\x0e\x32\x0c.RequestType\x12&\n\x07\x64\x65tails\x18\x02 \x01(\x0b\x32\x15.TxRequestDetailsType\x12,\n\nserialized\x18\x03 \x01(\x0b\x32\x18.TxRequestSerializedType\"%\n\x05TxAck\x12\x1c\n\x02tx\x18\x01 \x01(\x0b\x32\x10.TransactionType\"+\n\x08RawTxAck\x12\x1f\n\x02tx\x18\x01 \x01(\x0b\x32\x13.RawTransactionType\"}\n\x0cSignIdentity\x12\x1f\n\x08identity\x18\x01 \x01(\x0b\x32\r.IdentityType\x12\x18\n\x10\x63hallenge_hidden\x18\x02 \x01(\x0c\x12\x18\n\x10\x63hallenge_visual\x18\x03 \x01(\t\x12\x18\n\x10\x65\x63\x64sa_curve_name\x18\x04 \x01(\t\"H\n\x0eSignedIdentity\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\t\x12\x12\n\npublic_key\x18\x02 \x01(\x0c\x12\x11\n\tsignature\x18\x03 \x01(\x0c\",\n\rApplyPolicies\x12\x1b\n\x06policy\x18\x01 \x03(\x0b\x32\x0b.PolicyType\"?\n\tFlashHash\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\x0e\n\x06length\x18\x02 \x01(\r\x12\x11\n\tchallenge\x18\x03 \x01(\x0c\":\n\nFlashWrite\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\x0c\n\x04\x64\x61ta\x18\x02 \x01(\x0c\x12\r\n\x05\x65rase\x18\x03 \x01(\x08\"!\n\x11\x46lashHashResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"5\n\x12\x44\x65\x62ugLinkFlashDump\x12\x0f\n\x07\x61\x64\x64ress\x18\x01 \x01(\r\x12\x0e\n\x06length\x18\x02 \x01(\r\"*\n\x1a\x44\x65\x62ugLinkFlashDumpResponse\x12\x0c\n\x04\x64\x61ta\x18\x01 \x01(\x0c\"\x0b\n\tSoftReset\"\x0f\n\rFirmwareErase\"7\n\x0e\x46irmwareUpload\x12\x14\n\x0cpayload_hash\x18\x01 \x02(\x0c\x12\x0f\n\x07payload\x18\x02 \x02(\x0c\"#\n\x11\x44\x65\x62ugLinkDecision\x12\x0e\n\x06yes_no\x18\x01 \x02(\x08\"\x13\n\x11\x44\x65\x62ugLinkGetState\"\xd7\x02\n\x0e\x44\x65\x62ugLinkState\x12\x0e\n\x06layout\x18\x01 \x01(\x0c\x12\x0b\n\x03pin\x18\x02 \x01(\t\x12\x0e\n\x06matrix\x18\x03 \x01(\t\x12\x10\n\x08mnemonic\x18\x04 \x01(\t\x12\x19\n\x04node\x18\x05 \x01(\x0b\x32\x0b.HDNodeType\x12\x1d\n\x15passphrase_protection\x18\x06 \x01(\x08\x12\x12\n\nreset_word\x18\x07 \x01(\t\x12\x15\n\rreset_entropy\x18\x08 \x01(\x0c\x12\x1a\n\x12recovery_fake_word\x18\t \x01(\t\x12\x19\n\x11recovery_word_pos\x18\n \x01(\r\x12\x17\n\x0frecovery_cipher\x18\x0b \x01(\t\x12$\n\x1crecovery_auto_completed_word\x18\x0c \x01(\t\x12\x15\n\rfirmware_hash\x18\r \x01(\x0c\x12\x14\n\x0cstorage_hash\x18\x0e \x01(\x0c\"\x0f\n\rDebugLinkStop\";\n\x0c\x44\x65\x62ugLinkLog\x12\r\n\x05level\x18\x01 \x01(\r\x12\x0e\n\x06\x62ucket\x18\x02 \x01(\t\x12\x0c\n\x04text\x18\x03 \x01(\t\"\x15\n\x13\x44\x65\x62ugLinkFillConfig*\x88\x12\n\x0bMessageType\x12 \n\x16MessageType_Initialize\x10\x00\x1a\x04\x90\xb5\x18\x01\x12\x1a\n\x10MessageType_Ping\x10\x01\x1a\x04\x90\xb5\x18\x01\x12\x1d\n\x13MessageType_Success\x10\x02\x1a\x04\x98\xb5\x18\x01\x12\x1d\n\x13MessageType_Failure\x10\x03\x1a\x04\x98\xb5\x18\x01\x12\x1f\n\x15MessageType_ChangePin\x10\x04\x1a\x04\x90\xb5\x18\x01\x12 \n\x16MessageType_WipeDevice\x10\x05\x1a\x04\x90\xb5\x18\x01\x12#\n\x19MessageType_FirmwareErase\x10\x06\x1a\x04\x90\xb5\x18\x01\x12$\n\x1aMessageType_FirmwareUpload\x10\x07\x1a\x04\x90\xb5\x18\x01\x12 \n\x16MessageType_GetEntropy\x10\t\x1a\x04\x90\xb5\x18\x01\x12\x1d\n\x13MessageType_Entropy\x10\n\x1a\x04\x98\xb5\x18\x01\x12\"\n\x18MessageType_GetPublicKey\x10\x0b\x1a\x04\x90\xb5\x18\x01\x12\x1f\n\x15MessageType_PublicKey\x10\x0c\x1a\x04\x98\xb5\x18\x01\x12 \n\x16MessageType_LoadDevice\x10\r\x1a\x04\x90\xb5\x18\x01\x12!\n\x17MessageType_ResetDevice\x10\x0e\x1a\x04\x90\xb5\x18\x01\x12\x1c\n\x12MessageType_SignTx\x10\x0f\x1a\x04\x90\xb5\x18\x01\x12\x1e\n\x14MessageType_Features\x10\x11\x1a\x04\x98\xb5\x18\x01\x12&\n\x1cMessageType_PinMatrixRequest\x10\x12\x1a\x04\x98\xb5\x18\x01\x12\"\n\x18MessageType_PinMatrixAck\x10\x13\x1a\x04\x90\xb5\x18\x01\x12\x1c\n\x12MessageType_Cancel\x10\x14\x1a\x04\x90\xb5\x18\x01\x12\x1f\n\x15MessageType_TxRequest\x10\x15\x1a\x04\x98\xb5\x18\x01\x12\x1b\n\x11MessageType_TxAck\x10\x16\x1a\x04\x90\xb5\x18\x01\x12$\n\x1aMessageType_CipherKeyValue\x10\x17\x1a\x04\x90\xb5\x18\x01\x12\"\n\x18MessageType_ClearSession\x10\x18\x1a\x04\x90\xb5\x18\x01\x12#\n\x19MessageType_ApplySettings\x10\x19\x1a\x04\x90\xb5\x18\x01\x12#\n\x19MessageType_ButtonRequest\x10\x1a\x1a\x04\x98\xb5\x18\x01\x12\x1f\n\x15MessageType_ButtonAck\x10\x1b\x1a\x04\x90\xb5\x18\x01\x12 \n\x16MessageType_GetAddress\x10\x1d\x1a\x04\x90\xb5\x18\x01\x12\x1d\n\x13MessageType_Address\x10\x1e\x1a\x04\x98\xb5\x18\x01\x12$\n\x1aMessageType_EntropyRequest\x10#\x1a\x04\x98\xb5\x18\x01\x12 \n\x16MessageType_EntropyAck\x10$\x1a\x04\x90\xb5\x18\x01\x12!\n\x17MessageType_SignMessage\x10&\x1a\x04\x90\xb5\x18\x01\x12#\n\x19MessageType_VerifyMessage\x10\'\x1a\x04\x90\xb5\x18\x01\x12&\n\x1cMessageType_MessageSignature\x10(\x1a\x04\x98\xb5\x18\x01\x12\'\n\x1dMessageType_PassphraseRequest\x10)\x1a\x04\x98\xb5\x18\x01\x12#\n\x19MessageType_PassphraseAck\x10*\x1a\x04\x90\xb5\x18\x01\x12$\n\x1aMessageType_RecoveryDevice\x10-\x1a\x04\x90\xb5\x18\x01\x12!\n\x17MessageType_WordRequest\x10.\x1a\x04\x98\xb5\x18\x01\x12\x1d\n\x13MessageType_WordAck\x10/\x1a\x04\x90\xb5\x18\x01\x12&\n\x1cMessageType_CipheredKeyValue\x10\x30\x1a\x04\x98\xb5\x18\x01\x12$\n\x1aMessageType_EncryptMessage\x10\x31\x1a\x04\x90\xb5\x18\x01\x12&\n\x1cMessageType_EncryptedMessage\x10\x32\x1a\x04\x98\xb5\x18\x01\x12$\n\x1aMessageType_DecryptMessage\x10\x33\x1a\x04\x90\xb5\x18\x01\x12&\n\x1cMessageType_DecryptedMessage\x10\x34\x1a\x04\x98\xb5\x18\x01\x12\"\n\x18MessageType_SignIdentity\x10\x35\x1a\x04\x90\xb5\x18\x01\x12$\n\x1aMessageType_SignedIdentity\x10\x36\x1a\x04\x98\xb5\x18\x01\x12!\n\x17MessageType_GetFeatures\x10\x37\x1a\x04\x90\xb5\x18\x01\x12&\n\x1cMessageType_CharacterRequest\x10P\x1a\x04\x98\xb5\x18\x01\x12\"\n\x18MessageType_CharacterAck\x10Q\x1a\x04\x90\xb5\x18\x01\x12\x1e\n\x14MessageType_RawTxAck\x10R\x1a\x04\x90\xb5\x18\x01\x12#\n\x19MessageType_ApplyPolicies\x10S\x1a\x04\x90\xb5\x18\x01\x12\x1f\n\x15MessageType_FlashHash\x10T\x1a\x04\x90\xb5\x18\x01\x12 \n\x16MessageType_FlashWrite\x10U\x1a\x04\x90\xb5\x18\x01\x12\'\n\x1dMessageType_FlashHashResponse\x10V\x1a\x04\x98\xb5\x18\x01\x12(\n\x1eMessageType_DebugLinkFlashDump\x10W\x1a\x04\xa0\xb5\x18\x01\x12\x30\n&MessageType_DebugLinkFlashDumpResponse\x10X\x1a\x04\xa8\xb5\x18\x01\x12\x1f\n\x15MessageType_SoftReset\x10Y\x1a\x04\xa0\xb5\x18\x01\x12\'\n\x1dMessageType_DebugLinkDecision\x10\x64\x1a\x04\xa0\xb5\x18\x01\x12\'\n\x1dMessageType_DebugLinkGetState\x10\x65\x1a\x04\xa0\xb5\x18\x01\x12$\n\x1aMessageType_DebugLinkState\x10\x66\x1a\x04\xa8\xb5\x18\x01\x12#\n\x19MessageType_DebugLinkStop\x10g\x1a\x04\xa0\xb5\x18\x01\x12\"\n\x18MessageType_DebugLinkLog\x10h\x1a\x04\xa8\xb5\x18\x01\x12)\n\x1fMessageType_DebugLinkFillConfig\x10i\x1a\x04\xa8\xb5\x18\x01\x12\"\n\x18MessageType_GetCoinTable\x10j\x1a\x04\x90\xb5\x18\x01\x12\x1f\n\x15MessageType_CoinTable\x10k\x1a\x04\x98\xb5\x18\x01\x42,\n\x1a\x63om.keepkey.deviceprotocolB\x0eKeepKeyMessage')

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'messages_pb2', globals())
if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  DESCRIPTOR._serialized_options = b'\n\032com.keepkey.deviceprotocolB\016KeepKeyMessage'
  _MESSAGETYPE.values_by_name["MessageType_Initialize"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Initialize"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Ping"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Ping"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Success"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Success"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Failure"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Failure"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ChangePin"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ChangePin"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_WipeDevice"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_WipeDevice"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_FirmwareErase"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_FirmwareErase"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_FirmwareUpload"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_FirmwareUpload"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_GetEntropy"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_GetEntropy"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Entropy"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Entropy"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_GetPublicKey"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_GetPublicKey"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_PublicKey"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_PublicKey"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_LoadDevice"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_LoadDevice"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ResetDevice"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ResetDevice"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_SignTx"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_SignTx"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Features"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Features"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_PinMatrixRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_PinMatrixRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_PinMatrixAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_PinMatrixAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Cancel"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Cancel"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_TxRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_TxRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_TxAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_TxAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_CipherKeyValue"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_CipherKeyValue"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ClearSession"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ClearSession"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ApplySettings"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ApplySettings"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ButtonRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ButtonRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ButtonAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ButtonAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_GetAddress"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_GetAddress"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_Address"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_Address"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_EntropyRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_EntropyRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_EntropyAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_EntropyAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_SignMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_SignMessage"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_VerifyMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_VerifyMessage"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_MessageSignature"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_MessageSignature"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_PassphraseRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_PassphraseRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_PassphraseAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_PassphraseAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_RecoveryDevice"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_RecoveryDevice"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_WordRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_WordRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_WordAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_WordAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_CipheredKeyValue"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_CipheredKeyValue"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_EncryptMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_EncryptMessage"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_EncryptedMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_EncryptedMessage"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DecryptMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DecryptMessage"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DecryptedMessage"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DecryptedMessage"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_SignIdentity"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_SignIdentity"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_SignedIdentity"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_SignedIdentity"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_GetFeatures"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_GetFeatures"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_CharacterRequest"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_CharacterRequest"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_CharacterAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_CharacterAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_RawTxAck"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_RawTxAck"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_ApplyPolicies"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_ApplyPolicies"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_FlashHash"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_FlashHash"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_FlashWrite"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_FlashWrite"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_FlashHashResponse"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_FlashHashResponse"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFlashDump"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFlashDump"]._serialized_options = b'\240\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFlashDumpResponse"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFlashDumpResponse"]._serialized_options = b'\250\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_SoftReset"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_SoftReset"]._serialized_options = b'\240\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkDecision"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkDecision"]._serialized_options = b'\240\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkGetState"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkGetState"]._serialized_options = b'\240\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkState"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkState"]._serialized_options = b'\250\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkStop"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkStop"]._serialized_options = b'\240\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkLog"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkLog"]._serialized_options = b'\250\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFillConfig"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_DebugLinkFillConfig"]._serialized_options = b'\250\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_GetCoinTable"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_GetCoinTable"]._serialized_options = b'\220\265\030\001'
  _MESSAGETYPE.values_by_name["MessageType_CoinTable"]._options = None
  _MESSAGETYPE.values_by_name["MessageType_CoinTable"]._serialized_options = b'\230\265\030\001'
  _MESSAGETYPE._serialized_start=4978
  _MESSAGETYPE._serialized_end=7290
  _INITIALIZE._serialized_start=31
  _INITIALIZE._serialized_end=43
  _GETFEATURES._serialized_start=45
  _GETFEATURES._serialized_end=58
  _FEATURES._serialized_start=61
  _FEATURES._serialized_end=557
  _GETCOINTABLE._serialized_start=559
  _GETCOINTABLE._serialized_end=601
  _COINTABLE._serialized_start=603
  _COINTABLE._serialized_end=679
  _CLEARSESSION._serialized_start=681
  _CLEARSESSION._serialized_end=695
  _APPLYSETTINGS._serialized_start=697
  _APPLYSETTINGS._serialized_end=818
  _CHANGEPIN._serialized_start=820
  _CHANGEPIN._serialized_end=847
  _PING._serialized_start=849
  _PING._serialized_end=954
  _SUCCESS._serialized_start=956
  _SUCCESS._serialized_end=982
  _FAILURE._serialized_start=984
  _FAILURE._serialized_end=1038
  _BUTTONREQUEST._serialized_start=1040
  _BUTTONREQUEST._serialized_end=1103
  _BUTTONACK._serialized_start=1105
  _BUTTONACK._serialized_end=1116
  _PINMATRIXREQUEST._serialized_start=1118
  _PINMATRIXREQUEST._serialized_end=1173
  _PINMATRIXACK._serialized_start=1175
  _PINMATRIXACK._serialized_end=1202
  _CANCEL._serialized_start=1204
  _CANCEL._serialized_end=1212
  _PASSPHRASEREQUEST._serialized_start=1214
  _PASSPHRASEREQUEST._serialized_end=1233
  _PASSPHRASEACK._serialized_start=1235
  _PASSPHRASEACK._serialized_end=1270
  _GETENTROPY._serialized_start=1272
  _GETENTROPY._serialized_end=1298
  _ENTROPY._serialized_start=1300
  _ENTROPY._serialized_end=1326
  _GETPUBLICKEY._serialized_start=1329
  _GETPUBLICKEY._serialized_end=1491
  _PUBLICKEY._serialized_start=1493
  _PUBLICKEY._serialized_end=1545
  _GETADDRESS._serialized_start=1548
  _GETADDRESS._serialized_end=1727
  _ADDRESS._serialized_start=1729
  _ADDRESS._serialized_end=1755
  _WIPEDEVICE._serialized_start=1757
  _WIPEDEVICE._serialized_end=1769
  _LOADDEVICE._serialized_start=1772
  _LOADDEVICE._serialized_end=1959
  _RESETDEVICE._serialized_start=1962
  _RESETDEVICE._serialized_end=2187
  _ENTROPYREQUEST._serialized_start=2189
  _ENTROPYREQUEST._serialized_end=2205
  _ENTROPYACK._serialized_start=2207
  _ENTROPYACK._serialized_end=2236
  _RECOVERYDEVICE._serialized_start=2239
  _RECOVERYDEVICE._serialized_end=2494
  _WORDREQUEST._serialized_start=2496
  _WORDREQUEST._serialized_end=2509
  _WORDACK._serialized_start=2511
  _WORDACK._serialized_end=2534
  _CHARACTERREQUEST._serialized_start=2536
  _CHARACTERREQUEST._serialized_end=2595
  _CHARACTERACK._serialized_start=2597
  _CHARACTERACK._serialized_end=2660
  _SIGNMESSAGE._serialized_start=2663
  _SIGNMESSAGE._serialized_end=2793
  _VERIFYMESSAGE._serialized_start=2795
  _VERIFYMESSAGE._serialized_end=2891
  _MESSAGESIGNATURE._serialized_start=2893
  _MESSAGESIGNATURE._serialized_end=2947
  _ENCRYPTMESSAGE._serialized_start=2949
  _ENCRYPTMESSAGE._serialized_end=3067
  _ENCRYPTEDMESSAGE._serialized_start=3069
  _ENCRYPTEDMESSAGE._serialized_end=3133
  _DECRYPTMESSAGE._serialized_start=3135
  _DECRYPTMESSAGE._serialized_end=3216
  _DECRYPTEDMESSAGE._serialized_start=3218
  _DECRYPTEDMESSAGE._serialized_end=3270
  _CIPHERKEYVALUE._serialized_start=3273
  _CIPHERKEYVALUE._serialized_end=3413
  _CIPHEREDKEYVALUE._serialized_start=3415
  _CIPHEREDKEYVALUE._serialized_end=3448
  _SIGNTX._serialized_start=3451
  _SIGNTX._serialized_end=3657
  _TXREQUEST._serialized_start=3660
  _TXREQUEST._serialized_end=3793
  _TXACK._serialized_start=3795
  _TXACK._serialized_end=3832
  _RAWTXACK._serialized_start=3834
  _RAWTXACK._serialized_end=3877
  _SIGNIDENTITY._serialized_start=3879
  _SIGNIDENTITY._serialized_end=4004
  _SIGNEDIDENTITY._serialized_start=4006
  _SIGNEDIDENTITY._serialized_end=4078
  _APPLYPOLICIES._serialized_start=4080
  _APPLYPOLICIES._serialized_end=4124
  _FLASHHASH._serialized_start=4126
  _FLASHHASH._serialized_end=4189
  _FLASHWRITE._serialized_start=4191
  _FLASHWRITE._serialized_end=4249
  _FLASHHASHRESPONSE._serialized_start=4251
  _FLASHHASHRESPONSE._serialized_end=4284
  _DEBUGLINKFLASHDUMP._serialized_start=4286
  _DEBUGLINKFLASHDUMP._serialized_end=4339
  _DEBUGLINKFLASHDUMPRESPONSE._serialized_start=4341
  _DEBUGLINKFLASHDUMPRESPONSE._serialized_end=4383
  _SOFTRESET._serialized_start=4385
  _SOFTRESET._serialized_end=4396
  _FIRMWAREERASE._serialized_start=4398
  _FIRMWAREERASE._serialized_end=4413
  _FIRMWAREUPLOAD._serialized_start=4415
  _FIRMWAREUPLOAD._serialized_end=4470
  _DEBUGLINKDECISION._serialized_start=4472
  _DEBUGLINKDECISION._serialized_end=4507
  _DEBUGLINKGETSTATE._serialized_start=4509
  _DEBUGLINKGETSTATE._serialized_end=4528
  _DEBUGLINKSTATE._serialized_start=4531
  _DEBUGLINKSTATE._serialized_end=4874
  _DEBUGLINKSTOP._serialized_start=4876
  _DEBUGLINKSTOP._serialized_end=4891
  _DEBUGLINKLOG._serialized_start=4893
  _DEBUGLINKLOG._serialized_end=4952
  _DEBUGLINKFILLCONFIG._serialized_start=4954
  _DEBUGLINKFILLCONFIG._serialized_end=4975
# @@protoc_insertion_point(module_scope)
